from torchnyan.assertion import assert_close
from torchnyan.assertion import assert_grad_close
from torchnyan.assertion import assert_sequence_close
from torchnyan.assertion import assert_sequence_close
from torchnyan.strategy import BATCH_SIZE
from torchnyan.strategy import FEATURE_DIM
from torchnyan.strategy import TINY_BATCH_SIZE
from torchnyan.strategy import TINY_FEATURE_DIM
from torchnyan.strategy import TINY_TOKEN_SIZE
from torchnyan.strategy import TOKEN_SIZE
from torchnyan.strategy import device
from torchnyan.strategy import sizes

__all__ = [
    'assert_close', 'assert_grad_close',
    'assert_sequence_close',
    'assert_sequence_close',

    'TINY_BATCH_SIZE', 'TINY_TOKEN_SIZE', 'TINY_FEATURE_DIM',
    'BATCH_SIZE', 'TOKEN_SIZE', 'FEATURE_DIM',
    'device',
    'sizes',
]



from typing import Tuple

import torch
from torch import Tensor
from torch.testing import assert_close


def assert_grad_close(actual: Tensor, expected: Tensor, inputs, **kwargs) -> None:
    grad = torch.randn_like(actual)

    actual_grads = torch.autograd.grad(actual, inputs, grad, retain_graph=True, allow_unused=False)
    expected_grads = torch.autograd.grad(expected, inputs, grad, retain_graph=True, allow_unused=False)

    for actual_grad, expected_grad in zip(actual_grads, expected_grads):
        assert_close(actual=actual_grad, expected=expected_grad, **kwargs)


def assert_sequence_close(actual: Tuple[Tensor, ...], expected: Tuple[Tensor, ...], **kwargs) -> None:
    assert type(actual) == type(expected), f'{type(actual)} != {type(expected)}'
    assert len(actual) == len(expected), f'{len(actual)} == {len(expected)}'

    for index, (actual, expected) in enumerate(zip(actual, expected)):
        if actual is None:
            assert expected is None
        else:
            assert_close(actual=actual, expected=expected, **kwargs)



import torch
from hypothesis import strategies as st
from torch.nn import init

TINY_BATCH_SIZE = 5
TINY_TOKEN_SIZE = 11
TINY_FEATURE_DIM = 13

if torch.cuda.is_available():
    BATCH_SIZE = 53
    TOKEN_SIZE = 83
    FEATURE_DIM = 107
else:
    BATCH_SIZE = 37
    TOKEN_SIZE = 53
    FEATURE_DIM = 61

if torch.cuda.is_available():
    device = torch.device('cuda:0')
else:
    device = torch.device('cpu')

tensor = torch.empty((1024, 1024), device=device)
init.orthogonal_(tensor)
del tensor


@st.composite
def sizes(draw, *size: int, min_size: int = 1):
    max_size, *size = size
    n = draw(st.integers(min_value=min_size, max_value=max_size))

    if len(size) == 0:
        return n

    return draw(st.lists(sizes(*size), min_size=n, max_size=n))
